#include <string>
#include <iostream>
#include <fstream>

#include "util.h"
#include "NN.h"
#include "NN_statistics.h"

using namespace std;

NN *NN_statistics::create_NN_based_on_type_enum(NN_type nn_type, std::string nnetConfigFile, std::string trainingDataFile, std::string algorithmConfigFile)
{
    NN *nn = NULL;
    switch(nn_type)
    {
    case NN_statistics::PSO_CPU:
        nn = new NN_PSO(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case PSO_OpenCL_local_GPU:
        nn = new NN_PSO_OpenCL_GPU_local(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case PSO_OpenCL_local_CPU:
        nn = new NN_PSO_OpenCL_CPU_local(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case PSO_OpenCL_private_GPU:
        nn = new NN_PSO_OpenCL_GPU_private(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case PSO_OpenCL_private_CPU:
        nn = new NN_PSO_OpenCL_CPU_private(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case BP_CPU:
        nn = new NN_BP(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case BP_OpenCL_GPU:
        nn = new NN_BP_OpenCL_GPU(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    case BP_OpenCL_CPU:
        nn = new NN_BP_OpenCL_CPU(nnetConfigFile, trainingDataFile, algorithmConfigFile);
        break;
    default:
        error("Wrong NN_type");
    }
    return nn;
}

void NN_statistics::destroy_NN_based_on_type_enum(NN_type nn_type, NN *nn)
{
    switch(nn_type)
    {
    case NN_statistics::PSO_CPU:
        delete dynamic_cast<NN_PSO *>(nn);
        break;
    case PSO_OpenCL_local_GPU:
        delete dynamic_cast<NN_PSO_OpenCL_GPU_local *>(nn);
        break;
    case PSO_OpenCL_local_CPU:
        delete dynamic_cast<NN_PSO_OpenCL_CPU_local *>(nn);
        break;
    case PSO_OpenCL_private_GPU:
        delete dynamic_cast<NN_PSO_OpenCL_GPU_private *>(nn);
        break;
    case PSO_OpenCL_private_CPU:
        delete dynamic_cast<NN_PSO_OpenCL_CPU_private *>(nn);
        break;
    case BP_CPU:
        delete dynamic_cast<NN_BP *>(nn);
        break;
    case BP_OpenCL_GPU:
        delete dynamic_cast<NN_BP_OpenCL_GPU *>(nn);
        break;
    case BP_OpenCL_CPU:
        delete dynamic_cast<NN_BP_OpenCL_CPU *>(nn);
        break;
    default:
        error("Wrong NN_type");
    }

}

void NN_statistics::NN_sample_size_timings(NN_type nn_type, int start, int increment, int end, string outputFile, int numIterations,
                                           string nnetConfigFile, string trainingDataFile, string algorithmConfigFile)
{
    ofstream file(outputFile.c_str());
    if(!file.is_open()) error("Error opening file (sample size stats)");

    NN *nn = NULL;
    nn = create_NN_based_on_type_enum(nn_type, nnetConfigFile, trainingDataFile, algorithmConfigFile);

    file << "samples; time (s)" << endl;

    for(int i=end;i>=start;i-=increment)
    {
        cout << "i = " << i << endl;
        nn->downsizeTrainingSamples(i);
        double time = nn->timing(numIterations);
        //delete nn;
        destroy_NN_based_on_type_enum(nn_type, nn);
        file << i << ";" << time << "\n";
    }

    file.close();
}

void NN_statistics::NN_num_neurons_in_layer_timings(NN_type nn_type, int start, int increment, int end, std::string outputFile, int numIterations,
                                       int numInputs, int numOutputs, std::string trainingDataFile, std::string algorithmConfigFile, int numSamples, int numHiddenLayers)
{
    string nnConfigFileName = "nn_config_file_neur_in_layer_timing.txt";
    vector<int> layers;
    layers.push_back(numInputs);
    for(int i=0;i<numHiddenLayers;i++) layers.push_back(start);
    layers.push_back(numOutputs);


    ofstream output(outputFile.c_str());
    if(!output.is_open()) error("Error opening output file file (num neurons in layer timings)");
    output << "neurons in layer;time (s)\n";

    for(int i=start; i<=end; i+=increment)
    {
        cout << "i=" << i << " ";
        for(int j=0;j<numHiddenLayers;j++) layers[j+1] = i; //layers[1] = i;

        // create file
        ofstream nnConfigFile(nnConfigFileName.c_str());
        if(!nnConfigFile.is_open()) error("Error opening config file (num neurons in layer timings)");
        for(size_t j=0;j<layers.size();j++) nnConfigFile << layers[j] << " ";
        nnConfigFile.close();

        NN *nn = NULL;
        nn = create_NN_based_on_type_enum(nn_type,nnConfigFileName, trainingDataFile, algorithmConfigFile);
        nn->downsizeTrainingSamples(numSamples);
        double time = nn->timing(numIterations);
        output << i << ";" << time << endl;
        cout << time << "   total weights=" << nn->getNumWeights() << endl;
        //delete nn;
        destroy_NN_based_on_type_enum(nn_type, nn);


    }

    output.close();
}

void NN_statistics::NN_num_layers_timings(NN_type nn_type, int start, int increment, int end, int neuronsInLayer, std::string outputFile, int numIterations,
                                       int numInputs, int numOutputs, std::string trainingDataFile, std::string algorithmConfigFile, int numSamples)
{
    string nnConfigFileName = "nn_config_file_num_layers_timing.txt";
    vector<int> layers(2,0);
    layers[0] = numInputs; layers[1] = numOutputs;

    ofstream output(outputFile.c_str());
    if(!output.is_open()) error("Error opening output file file (num neurons in layer timings)");
    output << "layers;time (s)\n";

    for(int i=start; i<=end; i+=increment)
    {
        cout << "i=" << i << " ";

        // there have to be i hidden layers
        while((int)layers.size() < i + 2)
        {
            layers.back() = neuronsInLayer;
            layers.push_back(numOutputs);
        }

        // create file
        ofstream nnConfigFile(nnConfigFileName.c_str());
        if(!nnConfigFile.is_open()) error("Error opening config file (num layers timings)");
        for(size_t j=0;j<layers.size();j++) nnConfigFile << layers[j] << " ";
        nnConfigFile.close();

        NN *nn = NULL;
        nn = create_NN_based_on_type_enum(nn_type,nnConfigFileName, trainingDataFile, algorithmConfigFile);
        nn->downsizeTrainingSamples(numSamples);
        double time = nn->timing(numIterations);
        //delete nn;
        destroy_NN_based_on_type_enum(nn_type, nn);
        output << i << ";" << time << endl;
        cout << time << endl;
    }

    output.close();
}

void NN_statistics::Job_Scheduling_statistics(std::string fileNameExtension,std::string nnetConfigFile, std::string psoConfigFile)
{
    vector<string> trainingFiles(7);
    trainingFiles[0] = "1_CTC-SP2-1996-2.1-cln.swf" + fileNameExtension;
    trainingFiles[1] = "2_LANL-CM5-1994-3.1-cln.swf" + fileNameExtension;
    trainingFiles[2] = "3_LLNL-Atlas-2006-1.1-cln.swf" + fileNameExtension;
    trainingFiles[3] = "4_LLNL-Thunder-2007-1.1-cln.swf" + fileNameExtension;
    trainingFiles[4] = "5_LPC-EGEE-2004-1.2-cln.swf" + fileNameExtension;
    trainingFiles[5] = "6_SDSC-Par-1995-2.1-cln.swf" + fileNameExtension;
    trainingFiles[6] = "7_SDSC-Par-1996-2.1-cln.swf" + fileNameExtension;

    for(int i=0;i<7;i++)
    {
        cout << "Training file = " << trainingFiles[i] << endl;
        NN_PSO_OpenCL_GPU_local nn(nnetConfigFile, trainingFiles[i], psoConfigFile);
        double elapsed = nn.train();
        cout << "Elapsed = " << elapsed << " s" << endl;
        cout << "Test set MSE= " << nn.testMSE() << endl;
        cout << "Test set TPR= " << nn.testTPR() << endl;

        nn.makeLiftChart(trainingFiles[i] + "lift chart.cvs");
        nn.saveWeightsToFile(trainingFiles[i] + "weights.wgh");
    }
}


